from typing import Set, List, Optional

from .status import Status
from ..game import Game

from ..bancho.streams import StreamOut
from ..bancho.constants import ClientPackets, Privileges, Mode

import logging


class Player:
    def __init__(self, id: int, name: str = "", game: Optional[Game] = None) -> None:
        self.id = id
        self.name = name
        self.game = game

        self.timezone = 0
        self.country_code = 0
        self.mode = Mode.Osu
        self.longitude = 0.0
        self.latitude = 0.0

        self.status = Status()
        self.rscore = 0
        self.acc = 100.0
        self.playcount = 0
        self.tscore = 0
        self.rank = 0
        self.pp = 0

        self.privileges: Privileges = Privileges.Normal
        self.spectators: Set[Player] = set()

        self.cant_spectate = False
        self.silenced = False
        self.dms_blocked = False

        self.last_status = Status()

        self.logger = logging.getLogger(self.name)

    def __repr__(self) -> str:
        return f'<Player "{self.name}" ({self.id})>'

    def __hash__(self) -> int:
        return self.id

    def __eq__(self, other: object) -> bool:
        return self.id == other.id

    @property
    def loaded(self) -> bool:
        return bool(self.name)

    def request_presence(self):
        self.game.bancho.request_presence([self.id])

    def request_stats(self):
        self.game.bancho.request_stats([self.id])

    def avatar(self) -> Optional[bytes]:
        return self.game.api.get_avatar(self.id)

    def send_message(self, message: str):
        """Send a PM to this player"""

        if not (player := self.game.bancho.player):
            return

        if not self.loaded:
            # Presence missing
            self.request_presence()

        stream = StreamOut()
        stream.string(player.name)
        stream.string(message)
        stream.string(self.name)
        stream.s32(player.id)

        self.logger.info(
            f'<{player.name}{f" ({player.id})" if player.id else ""}> [{self.name}]: "{message}"'
        )

        self.game.bancho.enqueue(ClientPackets.SEND_PRIVATE_MESSAGE, stream.get())

    def add_friend(self):
        if self.id in self.game.bancho.friends:
            self.game.logger.warning(
                f"Tried to add friend, but you are already friends with {self.name}."
            )
            return

        self.game.logger.info(f"You are now friends with {self.name}.")

        self.game.bancho.enqueue(
            ClientPackets.FRIEND_ADD, int(self.id).to_bytes(4, "little")
        )

    def remove_friend(self):
        if self.id not in self.game.bancho.friends:
            self.game.logger.warning(
                f"Tried to remove friend, but you are not friends with {self.name}."
            )
            return

        self.game.logger.info(f"You are no longer friends with {self.name}.")

        self.game.bancho.enqueue(
            ClientPackets.FRIEND_REMOVE, int(self.id).to_bytes(4, "little")
        )



from typing import List

import struct


class StreamOut:
    def __init__(self, endian="<"):
        self.endian = endian
        self.data = bytearray()
        self.pos = 0
        self.stack = []

    def push(self):
        self.stack.append(self.pos)

    def pop(self):
        self.pos = self.stack.pop()

    def get(self):
        return bytes(self.data)

    def size(self):
        return len(self.data)

    def tell(self):
        return self.pos

    def seek(self, pos):
        if pos > len(self.data):
            self.data += bytes(pos - len(self.data))
        self.pos = pos

    def skip(self, num):
        self.seek(self.pos + num)

    def align(self, num):
        self.skip((num - self.pos % num) % num)

    def available(self):
        return len(self.data) - self.pos

    def eof(self):
        return self.pos >= len(self.data)

    def write(self, data):
        self.data[self.pos : self.pos + len(data)] = data
        self.pos += len(data)

    def pad(self, num, char=b"\0"):
        self.write(char * num)

    def ascii(self, data):
        self.write(data.encode("ascii"))

    def u8(self, value):
        self.write(bytes([value]))

    def u16(self, value):
        self.write(struct.pack(self.endian + "H", value))

    def u32(self, value):
        self.write(struct.pack(self.endian + "I", value))

    def u64(self, value):
        self.write(struct.pack(self.endian + "Q", value))

    def s8(self, value):
        self.write(struct.pack("b", value))

    def s16(self, value):
        self.write(struct.pack(self.endian + "h", value))

    def s32(self, value):
        self.write(struct.pack(self.endian + "i", value))

    def s64(self, value):
        self.write(struct.pack(self.endian + "q", value))

    def u24(self, value):
        if self.endian == ">":
            self.u16(value >> 8)
            self.u8(value & 0xFF)
        else:
            self.u8(value & 0xFF)
            self.u16(value >> 8)

    def float(self, value):
        self.write(struct.pack(self.endian + "f", value))

    def double(self, value):
        self.write(struct.pack(self.endian + "d", value))

    def bool(self, value):
        self.u8(1 if value else 0)

    def char(self, value):
        self.u8(ord(value))

    def wchar(self, value):
        self.u16(ord(value))

    def intlist(self, numbers: List[int]):
        self.s16(len(numbers))
        [self.s32(num) for num in numbers]

    def uleb128(self, value: int):
        if value == 0:
            return b"\x00"

        ret = bytearray()

        while value != 0:
            ret.append(value & 0x7F)
            value >>= 7
            if value != 0:
                ret[-1] |= 0x80

        self.write(bytes(ret))

    def string(self, value: str):
        string = value.encode()
        length = len(string)  # + 1

        if not string:
            self.s8(0x00)
            return

        self.s8(0x0B)
        self.uleb128(length)
        self.write(string)


class StreamIn:
    def __init__(self, data: bytes, endian="<"):
        self.endian = endian
        self.data = data
        self.pos = 0
        self.stack = []

    def push(self):
        self.stack.append(self.pos)

    def pop(self):
        self.pos = self.stack.pop()

    def get(self):
        return self.data

    def size(self):
        return len(self.data)

    def tell(self):
        return self.pos

    def seek(self, pos):
        if pos > self.size():
            raise OverflowError("Buffer overflow")
        self.pos = pos

    def skip(self, num):
        self.seek(self.pos + num)

    def align(self, num):
        self.skip((num - self.pos % num) % num)

    def eof(self):
        return self.pos == len(self.data)

    def available(self):
        return len(self.data) - self.pos

    def peek(self, num):
        if self.available() < num:
            raise OverflowError("Buffer overflow")
        return self.data[self.pos : self.pos + num]

    def read(self, num):
        data = self.peek(num)
        self.skip(num)
        return data

    def readall(self):
        return self.read(self.available())

    def pad(self, num, char=b"\0"):
        if (self.read(num)) != char * num:
            raise ValueError("Incorrect padding")

    def ascii(self, num):
        return self.read(num).decode("ascii")

    def u8(self):
        return self.read(1)[0]

    def u16(self):
        return struct.unpack(self.endian + "H", self.read(2))[0]

    def u32(self):
        return struct.unpack(self.endian + "I", self.read(4))[0]

    def u64(self):
        return struct.unpack(self.endian + "Q", self.read(8))[0]

    def s8(self):
        return struct.unpack("b", self.read(1))[0]

    def s16(self):
        return struct.unpack(self.endian + "h", self.read(2))[0]

    def s32(self):
        return struct.unpack(self.endian + "i", self.read(4))[0]

    def s64(self):
        return struct.unpack(self.endian + "q", self.read(8))[0]

    def u24(self):
        if self.endian == ">":
            return (self.u16() << 8) | self.u8()
        return self.u8() | (self.u16() << 8)

    def float(self):
        return struct.unpack(self.endian + "f", self.read(4))[0]

    def double(self):
        return struct.unpack(self.endian + "d", self.read(8))[0]

    def bool(self):
        return bool(self.u8())

    def char(self):
        return chr(self.u8())

    def wchar(self):
        return chr(self.u16())

    def uleb128(self) -> int:
        length = shift = 0

        while True:
            byte = self.u8()

            length |= (byte & 0x7F) << shift
            if (byte & 0x80) == 0:
                break

            shift += 7

        return length

    def string(self) -> str:
        empty = self.s8() == 0x00

        if empty:
            return ""

        size = self.uleb128()
        return self.read(size).decode()

    def encoded_string(self) -> bytes:
        empty = self.s8() == 0x00

        if empty:
            return ""

        size = self.uleb128()
        return self.read(size)

    def intlist(self) -> List[int]:
        return [self.s32() for num in range(self.s16())]



from concurrent.futures import ThreadPoolExecutor, Future
from typing import List, Dict, Callable

from .bancho.constants import ServerPackets


class EventHandler:

    """EventHandler
    ---------------

    This class is used to make it easier to handle packets.

    After a packet has been received from the server,
    it will call every event associated with that packet.

    Example of how a event can be registered and used:
    >>> game = Game(...)
    >>>
    >>> @game.events.register(ServerPackets.SEND_MESSAGE)
    >>> def message_handler(sender, message, target):
    >>>     print(message)
    """

    def __init__(self) -> None:
        self.handlers: Dict[ServerPackets, List[Callable]] = {}
        self.executor = ThreadPoolExecutor(max_workers=10)

    def register(self, packet: ServerPackets, threaded: bool = False):
        def wrapper(f: Callable):
            if threaded:
                f = self._submit_future(f)
            if packet in self.handlers:
                self.handlers[packet].append(f)
            else:
                self.handlers[packet] = [f]
            return f

        return wrapper

    def call(self, packet: ServerPackets, *args):
        if packet in self.handlers:
            for handler in self.handlers[packet]:
                handler(*args)

    def _submit_future(self, f: Callable) -> Callable:
        def execute(*args):
            future = self.executor.submit(f, *args)
            future.add_done_callback(self._future_callback)
            return future

        return execute

    def _future_callback(self, future: Future):
        ...  # TODO



from osu.api.constants import CommentTarget
from dataclasses import dataclass


@dataclass
class Comment:
    time: int
    target: CommentTarget
    format: str
    text: str

    @classmethod
    def from_string(cls, string: str):
        line = string.split("\t")
        return Comment(int(line[0]), CommentTarget(line[1]), line[2], line[3])



from ..bancho.constants import ClientPackets
from ..bancho.streams import StreamOut
from ..game import Game

from typing import Optional

import logging


class Channel:
    def __init__(self, name: str, game: Game, topic: Optional[str] = None) -> None:
        self.name = name
        self.topic = topic
        self.game = game
        self.user_count = 0

        self.joined = False
        self.joining = False

        self.logger = logging.getLogger(self.name)

    def __hash__(self) -> int:
        return int(self.name.encode().hex(), 16)

    def __eq__(self, other: object) -> bool:
        return self.name == other.name

    def join(self):
        if self.joined:
            return

        self.joining = True
        self.logger.info("Attempting to join channel...")

        stream = StreamOut()
        stream.string(self.name)

        self.game.bancho.enqueue(ClientPackets.CHANNEL_JOIN, stream.get())

    def leave(self):
        if not self.joined:
            return

        self.joining = False
        self.joined = False

        stream = StreamOut()
        stream.string(self.name)

        self.game.bancho.enqueue(ClientPackets.CHANNEL_PART, stream.get())

    def join_success(self):
        if not self.joined:
            self.logger.info(f"Joined {self.name}!")

        self.joining = False
        self.joined = True

        if self.name == "#osu":
            # Load players
            self.game.bancho.players.load()

    def send_message(self, message: str, force=False):
        if not self.joined and not force:
            return

        if not (player := self.game.bancho.player):
            return

        self.logger.info(
            f'<{player.name}{f" ({player.id})" if player.id else ""}> [{self.name}]: "{message}"'
        )

        stream = StreamOut()
        stream.string(player.name)
        stream.string(message)
        stream.string(self.name)
        stream.s32(player.id)

        self.game.bancho.enqueue(ClientPackets.SEND_PUBLIC_MESSAGE, stream.get())



from .replays import ReplayFrame, ScoreFrame
from .client import ClientHash, ClientInfo
from .beatmap import BeatmapInfo
from .channel import Channel
from .status import Status
from .player import Player



from ..api.constants import SubmissionStatus, Mods, Mode
from ..game import Game

from typing import Optional, List
from dataclasses import dataclass
from datetime import datetime


@dataclass
class Score:
    id: int
    username: str
    total_score: int
    max_combo: int
    c50: int
    c100: int
    c300: int
    cMiss: int
    cKatu: int
    cGeki: int
    perfect: bool
    mods: Mods
    user_id: int
    rank: Optional[int]
    date: Optional[datetime]
    mode: Mode
    has_replay: bool

    @classmethod
    def from_string(cls, string: str, mode: Mode):
        lines = string.split("|")

        return Score(
            int(lines[0]),
            lines[1],
            int(lines[2]),
            int(lines[3]),
            int(lines[4]),
            int(lines[5]),
            int(lines[6]),
            int(lines[7]),
            int(lines[8]),
            int(lines[9]),
            lines[10] == "1",
            Mods(int(lines[11])),
            int(lines[12]),
            int(lines[13]) if len(lines[13]) > 0 else 0,
            datetime.fromtimestamp(int(lines[14])),
            mode,
            lines[15] == "1",
        )

    def get_comments(self, game: Game):
        return game.api.get_comments(replay_id=self.id, mode=self.mode)

    def get_replay(self, game: Game):
        return game.api.get_replay(replay_id=self.id, mode=self.mode)


@dataclass
class ScoreResponse:
    status: SubmissionStatus
    beatmap_id: Optional[int]
    set_id: Optional[int]
    total_scores: int
    global_offset: int
    beatmap_format: Optional[str]
    rating: float
    personal_best: Optional[Score]
    scores: List[Score]

    @classmethod
    def from_string(cls, string: str, mode: Mode):
        result = string.split("\n")

        if len(result) <= 0:
            return

        status_results = result[0].split("|")

        beatmap_status = {
            "-1": SubmissionStatus.NotSubmitted,
            "0": SubmissionStatus.Pending,
            "1": SubmissionStatus.Unknown,
            "2": SubmissionStatus.Ranked,
            "3": SubmissionStatus.Approved,
            "4": SubmissionStatus.Qualified,
        }[status_results[0]]

        if len(status_results) > 2:
            beatmap_id = int(status_results[2])
            beatmapset_id = int(status_results[3])
            total_scores = int(status_results[4])
        else:
            beatmap_id = None
            beatmapset_id = None
            total_scores = 0

        if len(result) > 1:
            offset = int(result[1])
            beatmap_format = result[2]
            rating = float(result[3])
        else:
            offset = 0
            beatmap_format = None
            rating = 0.0

        personal_best = None
        scores = []

        if len(result) > 4:
            if len(result[4]) > 0:
                personal_best = Score.from_string(result[4])

            for score in result[5:]:
                if not score:
                    continue

                scores.append(Score.from_string(score, mode))

        return ScoreResponse(
            beatmap_status,
            beatmap_id,
            beatmapset_id,
            total_scores,
            offset,
            beatmap_format,
            rating,
            personal_best,
            scores,
        )



from dataclasses import dataclass
from typing import List

from ..bancho.streams import StreamIn, StreamOut
from ..bancho.constants import ButtonState


@dataclass
class ReplayFrame:
    button_state: ButtonState
    time: int
    x: float
    y: float

    def encode(self) -> bytes:
        stream = StreamOut()
        stream.u8(self.button_state.value)
        stream.u8(0)  # "Taiko-Byte"
        stream.float(self.x)
        stream.float(self.y)
        stream.s32(self.time)
        return stream.get()

    @classmethod
    def decode(cls, stream: StreamIn):
        button_state = ButtonState(stream.u8())

        # "Taiko-Byte"
        if stream.u8() > 0:
            if ButtonState.Right1 not in button_state:
                button_state |= ButtonState.Right1

        x = stream.float()
        y = stream.float()
        time = stream.s32()

        return ReplayFrame(button_state, time, x, y)


@dataclass
class ScoreFrame:
    time: int
    id: int
    c300: int
    c100: int
    c50: int
    cGeki: int
    cKatu: int
    cMiss: int
    total_score: int
    max_combo: int
    current_combo: int
    perfect: bool
    current_hp: int
    tag_byte: int
    score_v2: bool = False
    combo_portion: float = 0.0
    bonus_portion: float = 0.0

    @property
    def total_hits(self) -> int:
        return self.c50 + self.c100 + self.c300 + self.cMiss

    def encode(self) -> bytes:
        stream = StreamOut()
        stream.s32(self.time)
        stream.u8(self.id)
        stream.u16(self.c300)
        stream.u16(self.c100)
        stream.u16(self.c50)
        stream.u16(self.cGeki)
        stream.u16(self.cKatu)
        stream.u16(self.cMiss)
        stream.s32(self.total_score)
        stream.u16(self.max_combo)
        stream.u16(self.current_combo)
        stream.bool(self.perfect)
        stream.u8(self.current_hp)
        stream.u8(self.tag_byte)
        stream.bool(self.score_v2)
        if self.score_v2:
            stream.float(self.combo_portion)
            stream.float(self.bonus_portion)
        return stream.get()

    @classmethod
    def decode(cls, stream: StreamIn):
        return ScoreFrame(
            stream.s32(),
            stream.u8(),
            stream.u16(),
            stream.u16(),
            stream.u16(),
            stream.u16(),
            stream.u16(),
            stream.u16(),
            stream.s32(),
            stream.u16(),
            stream.u16(),
            stream.bool(),
            stream.u8(),
            stream.u8(),
            v2 := stream.bool(),
            stream.float() if v2 else 0.0,
            stream.float() if v2 else 0.0,
        )



from typing import List, Dict, Callable
from copy import copy

from ..objects.replays import ScoreFrame, ReplayFrame
from ..objects.beatmap import BeatmapInfo
from ..objects.channel import Channel
from ..objects.player import Player
from ..game import Game

from .streams import StreamIn
from .constants import (
    ServerPackets,
    ReplayAction,
    StatusAction,
    LoginError,
    Privileges,
    Mode,
    Mods,
)

import threading
import time


class PacketHandler:
    def __init__(self) -> None:
        self.handlers: Dict[ServerPackets, List[Callable]] = {}

    def register(self, packet: ServerPackets):
        def wrapper(f: Callable):
            if packet in self.handlers:
                self.handlers[packet].append(f)
            else:
                self.handlers[packet] = [f]
            return f

        return wrapper

    def data_received(self, data: bytes, game: Game):
        stream = StreamIn(data)

        while not stream.eof():
            packet = ServerPackets(stream.u16())
            compression = stream.bool()  # unused
            data = stream.read(stream.u32())

            game.logger.debug(f'Received packet {packet.name} -> "{data}"')

            # Handling packet
            self.packet_received(packet, StreamIn(data), game)

            # Reset stream
            stream = StreamIn(stream.readall())

    def packet_received(self, packet: ServerPackets, data: StreamIn, game: Game):
        if packet in self.handlers:
            for handler in self.handlers[packet]:
                try:
                    # TODO: Add threading
                    handler(data, game)
                except Exception as exc:
                    game.logger.error(
                        f"Error while executing {handler}: {exc}", exc_info=exc
                    )
        else:
            game.logger.warning(f'No handler found for "{packet.name}"')


Packets = PacketHandler()


@Packets.register(ServerPackets.PROTOCOL_VERSION)
def protocol_version(stream: StreamIn, game: Game):
    game.bancho.protocol = stream.s32()
    game.events.call(ServerPackets.PROTOCOL_VERSION, game.bancho.protocol)


@Packets.register(ServerPackets.RESTART)
def server_restart(stream: StreamIn, game: Game):
    timeout = stream.s32() / 1000

    game.logger.warning(f"Bancho is restarting. Retrying in {timeout} seconds...")
    game.bancho.connected = False
    game.bancho.retry = True

    game.events.call(ServerPackets.RESTART, game.bancho.protocol)

    time.sleep(timeout)


@Packets.register(ServerPackets.USER_ID)
def login_reply(stream: StreamIn, game: Game):
    response = stream.s32()

    if response < 0:
        # Got a login error
        try:
            error = LoginError(response)
        except ValueError:
            error = None

        game.logger.error(f"Login error: {error.name}")
        game.logger.error(error.description)
        game.bancho.connected = False
        game.bancho.retry = False

        if error == LoginError.SERVER_ERROR:
            game.bancho.retry = True

        elif error == LoginError.VERIFICATION_NEEDED:
            game.api.verify(game.client.hash)

        return

    game.logger.info(f"Logged in with id: {response}")
    game.bancho.user_id = response

    game.bancho.player = Player(response, game.username, game)
    game.bancho.players.add(game.bancho.player)

    game.bancho.fast_read = True

    game.events.call(ServerPackets.USER_ID, response)


@Packets.register(ServerPackets.PONG)
def pong(stream: StreamIn, game: Game):
    game.events.call(ServerPackets.PONG)


@Packets.register(ServerPackets.ACCOUNT_RESTRICTED)
def restricted(stream: StreamIn, game: Game):
    game.logger.error("You have been banned.")
    game.events.call(ServerPackets.ACCOUNT_RESTRICTED)
    game.bancho.connected = False
    game.bancho.retry = False


@Packets.register(ServerPackets.PRIVILEGES)
def privileges(stream: StreamIn, game: Game):
    game.bancho.privileges = Privileges(stream.s32())
    game.events.call(ServerPackets.PRIVILEGES, game.bancho.privileges)


@Packets.register(ServerPackets.FRIENDS_LIST)
def friends(stream: StreamIn, game: Game):
    game.bancho.friends = stream.intlist()
    game.events.call(ServerPackets.FRIENDS_LIST, game.bancho.friends)


@Packets.register(ServerPackets.MAIN_MENU_ICON)
def menu_icon(stream: StreamIn, game: Game):
    image, link = stream.string().split("|")
    game.events.call(ServerPackets.MAIN_MENU_ICON, image, link)


@Packets.register(ServerPackets.VERSION_UPDATE)
def version_update(stream: StreamIn, game: Game):
    game.logger.info("Bancho requested a version update.")
    game.events.call(ServerPackets.VERSION_UPDATE)


@Packets.register(ServerPackets.VERSION_UPDATE_FORCED)
def version_update_forced(stream: StreamIn, game: Game):
    game.logger.warning("Bancho forced a version update.")
    game.events.call(ServerPackets.VERSION_UPDATE_FORCED)


@Packets.register(ServerPackets.GET_ATTENTION)
def attension(stream: StreamIn, game: Game):
    game.events.call(ServerPackets.GET_ATTENTION)


@Packets.register(ServerPackets.NOTIFICATION)
def notification(stream: StreamIn, game: Game):
    message = stream.string()
    game.logger.info(f'Notification from bancho: "{message}"')
    game.events.call(ServerPackets.NOTIFICATION, message)


@Packets.register(ServerPackets.USER_PRESENCE)
def presence(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        # Add new player, if not found in collection
        game.bancho.players.add(player := Player(user_id, game=game))

    player.name = stream.string()
    player.timezone = stream.u8() - 24
    player.country_code = stream.u8()

    b = stream.u8()  # Contains privileges and play mode

    player.privileges = Privileges(b & -255)
    player.mode = Mode(max(0, min(3, (b & 224) >> 5)))

    player.longitude = stream.float()
    player.latitude = stream.float()
    player.rank = stream.s32()

    game.bancho.fast_read = True

    game.events.call(ServerPackets.USER_PRESENCE, player)


@Packets.register(ServerPackets.USER_STATS)
def stats(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        # Add new player, if not found in collection
        game.bancho.request_presence([user_id])
        game.bancho.players.add(player := Player(user_id, game=game))

    player.last_status = copy(player.status)

    # Status
    player.status.action = StatusAction(stream.u8())
    player.status.text = stream.string()
    player.status.checksum = stream.string()
    player.status.mods = Mods(stream.u32())
    player.status.mode = Mode(max(0, min(3, stream.u8())))
    player.status.beatmap_id = stream.s32()

    # Stats
    player.rscore = stream.s64()
    player.acc = stream.float()
    player.playcount = stream.s32()
    player.tscore = stream.s64()
    player.rank = stream.s32()
    player.pp = stream.s16()

    game.bancho.fast_read = True

    game.events.call(ServerPackets.USER_STATS, player)


@Packets.register(ServerPackets.USER_PRESENCE_BUNDLE)
def presence_bundle(stream: StreamIn, game: Game):
    user_ids = stream.intlist()

    for id in user_ids:
        if not (game.bancho.players.by_id(id)):
            # Add player if not found
            game.bancho.players.add(Player(id, game=game))

    game.bancho.fast_read = True

    game.events.call(ServerPackets.USER_PRESENCE_BUNDLE, user_ids)


@Packets.register(ServerPackets.USER_PRESENCE_SINGLE)
def presence_single(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (game.bancho.players.by_id(user_id)):
        # Add player if not found
        game.bancho.players.add(Player(user_id, game=game))

    game.events.call(ServerPackets.USER_PRESENCE_SINGLE, user_id)


@Packets.register(ServerPackets.USER_LOGOUT)
def logout(stream: StreamIn, game: Game):
    if not (player := game.bancho.players.by_id(stream.s32())):
        return

    if player is game.bancho.spectating:
        game.logger.info(f"Stopped spectating {player.name}.")
        game.bancho.spectating = None

    game.bancho.players.remove(player)

    game.events.call(ServerPackets.USER_LOGOUT, player)


@Packets.register(ServerPackets.SEND_MESSAGE)
def message(stream: StreamIn, game: Game):
    sender = stream.string()
    message = stream.string()
    target = stream.string()
    sender_id = stream.s32()

    # Find player
    if not (player := game.bancho.players.by_id(sender_id)):
        # Try to find sender by name
        if not (player := game.bancho.players.by_name(sender)):
            return

    if not player.loaded:
        # Presence missing
        player.request_presence()

    sender = player

    # Find target
    if target.startswith("#"):
        # Public message
        if not (channel := game.bancho.channels.get(target)):
            return

        target = channel
    else:
        # Private message
        if not (player := game.bancho.players.by_name(target)):
            return

        if not player.loaded:
            # Presence missing
            player.request_presence()

        target = player

    target.logger.info(
        f'<{sender.name}{f" ({sender_id})" if sender_id else ""}> [{target.name}]: "{message}"'
    )

    game.bancho.fast_read = True
    game.events.call(ServerPackets.SEND_MESSAGE, sender, message, target)


@Packets.register(ServerPackets.SPECTATOR_JOINED)
def spectator_joined(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.cant_spectate = False

    game.bancho.player.spectators.add(player)

    game.events.call(ServerPackets.SPECTATOR_JOINED, player)


@Packets.register(ServerPackets.SPECTATOR_LEFT)
def spectator_left(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    if player not in game.bancho.player.spectators:
        return

    game.bancho.player.spectators.remove(player)

    game.events.call(ServerPackets.SPECTATOR_LEFT, player)


@Packets.register(ServerPackets.FELLOW_SPECTATOR_JOINED)
def fellow_spectator_joined(stream: StreamIn, game: Game):
    if not game.bancho.spectating:
        return

    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    game.bancho.spectating.spectators.add(player)

    game.events.call(ServerPackets.FELLOW_SPECTATOR_JOINED, player)


@Packets.register(ServerPackets.FELLOW_SPECTATOR_LEFT)
def fellow_spectator_left(stream: StreamIn, game: Game):
    if not game.bancho.spectating:
        return

    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    if player not in game.bancho.spectating.spectators:
        return

    game.bancho.spectating.spectators.remove(player)

    game.events.call(ServerPackets.SPECTATOR_LEFT, player)


@Packets.register(ServerPackets.SPECTATE_FRAMES)
def frames(stream: StreamIn, game: Game):
    if not game.bancho.spectating:
        return

    extra = stream.s32()
    frames = [ReplayFrame.decode(stream) for _ in range(stream.u16())]
    action = ReplayAction(stream.u8())

    try:
        score_frame = ScoreFrame.decode(stream)
    except OverflowError:
        score_frame = None

    game.events.call(ServerPackets.SPECTATE_FRAMES, action, frames, score_frame, extra)


@Packets.register(ServerPackets.SPECTATOR_CANT_SPECTATE)
def cant_spectate(stream: StreamIn, game: Game):
    if not game.bancho.spectating:
        return

    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.cant_spectate = True

    game.events.call(ServerPackets.SPECTATOR_CANT_SPECTATE, player)


@Packets.register(ServerPackets.CHANNEL_INFO)
def channel_info(stream: StreamIn, game: Game):
    name = stream.string()
    topic = stream.string()
    user_count = stream.s16()

    if not (c := game.bancho.channels.get(name)):
        game.bancho.channels.add(c := Channel(name, game, topic))

    c.user_count = user_count
    c.topic = topic

    if c.name == "#osu" and not c.joined:
        c.join()

    game.events.call(ServerPackets.CHANNEL_INFO, c)


@Packets.register(ServerPackets.CHANNEL_AUTO_JOIN)
def channel_autojoin(stream: StreamIn, game: Game):
    name = stream.string()
    topic = stream.string()
    user_count = stream.s16()

    if not (c := game.bancho.channels.get(name)):
        game.bancho.channels.add(c := Channel(name, game, topic))

    c.user_count = user_count
    c.topic = topic
    c.join_success()

    game.events.call(ServerPackets.CHANNEL_AUTO_JOIN, c)


@Packets.register(ServerPackets.CHANNEL_INFO_END)
def channel_info_end(stream: StreamIn, game: Game):
    game.events.call(ServerPackets.CHANNEL_INFO_END)


@Packets.register(ServerPackets.CHANNEL_JOIN_SUCCESS)
def channel_join_success(stream: StreamIn, game: Game):
    name = stream.string()

    if not (c := game.bancho.channels.get(name)):
        game.bancho.channels.add(c := Channel(name=name, game=game))

    c.join_success()

    game.events.call(ServerPackets.CHANNEL_JOIN_SUCCESS, c)


@Packets.register(ServerPackets.CHANNEL_KICK)
def channel_revoked(stream: StreamIn, game: Game):
    name = stream.string()

    if not (c := game.bancho.channels.get(name)):
        return

    game.bancho.channels.remove(c)
    game.logger.info(f"Kicked out of channel: {name}")

    game.events.call(ServerPackets.CHANNEL_KICK, c)


@Packets.register(ServerPackets.BEATMAP_INFO_REPLY)
def beatmapinfo_reply(stream: StreamIn, game: Game):
    beatmaps = [BeatmapInfo.decode(stream) for beatmap in range(stream.s32())]

    game.events.call(ServerPackets.BEATMAP_INFO_REPLY, beatmaps)


@Packets.register(ServerPackets.SILENCE_END)
def silence_info(stream: StreamIn, game: Game):
    if (remaining_silence := stream.s32()) > 0:
        game.bancho.player.silenced = True
        game.bancho.silenced = True

        game.logger.warning(f"You have been silenced for {remaining_silence} seconds.")

        threading.Timer(remaining_silence, game.bancho.unsilence)
    else:
        game.bancho.unsilence()

    game.events.call(ServerPackets.SILENCE_END, remaining_silence)


@Packets.register(ServerPackets.USER_SILENCED)
def user_silenced(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.silenced = True

    game.logger.info(f"{player} has been silenced.")
    game.events.call(ServerPackets.USER_SILENCED, player)


@Packets.register(ServerPackets.TARGET_IS_SILENCED)
def target_silenced(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.silenced = True

    game.logger.info(
        f"{player.name} has been silenced and is unable to respond to your messages right now."
    )
    game.events.call(ServerPackets.TARGET_IS_SILENCED, player)


@Packets.register(ServerPackets.USER_DM_BLOCKED)
def dms_blocked(stream: StreamIn, game: Game):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.dms_blocked = True

    game.logger.info(f"{player} blocked their dms.")
    game.events.call(ServerPackets.USER_DM_BLOCKED, player)



from typing import List, Set, Iterator, Optional
from collections.abc import Iterator

from ..bancho.constants import ClientPackets, PresenceFilter
from ..game import Game

from .lists import LockedSet

from .channel import Channel
from .player import Player


class Players(LockedSet[Player]):
    def __init__(self, game: Game) -> None:
        self.game = game
        super().__init__()

    def __iter__(self) -> Iterator[Player]:
        return super().__iter__()

    @property
    def ids(self) -> Set[int]:
        return {p.id for p in self}

    @property
    def pending(self) -> List[Player]:
        return [p for p in self if not p.name]

    def add(self, player: Player) -> None:
        return super().add(player)

    def remove(self, player: Player) -> None:
        if player in self:
            return super().remove(player)

    def by_id(self, id: int) -> Optional[Player]:
        for p in self.copy():
            if p.id == id:
                return p
        return None

    def by_name(self, name: str) -> Optional[Player]:
        for p in self.copy():
            if p.name == name:
                return p
        return None

    def load(self):
        # Split players into chunks of 255
        player_chunks = [
            self.pending[i : i + 255] for i in range(0, len(self.pending), 255)
        ]

        for chunk in player_chunks:
            self.game.bancho.request_presence([p.id for p in chunk])

    def request_updates(self, filter=PresenceFilter.All):
        self.game.bancho.enqueue(
            ClientPackets.RECEIVE_UPDATES, int(filter.value).to_bytes(4, "little")
        )


class Channels(LockedSet[Channel]):
    def __iter__(self) -> Iterator[Channel]:
        return super().__iter__()

    @property
    def joined(self) -> List[Channel]:
        return [c for c in self if c.joined]

    def add(self, channel: Channel) -> None:
        return super().add(channel)

    def remove(self, channel: Channel) -> None:
        if channel in self:
            return super().remove(channel)

    def get(self, name: str) -> Optional[Channel]:
        for c in self.copy():
            if c.name == name:
                return c
        return None



from concurrent.futures import Future, ThreadPoolExecutor
from typing import List, Callable, Optional
from dataclasses import dataclass
from datetime import datetime
import logging


@dataclass
class Task:
    function: Callable
    interval: float
    loop: bool
    last_call: datetime
    threaded: bool


class TaskManager:
    """### TaskManager

    Used to run any task syncronously or asyncronously inside a thread.\n
    Example of registering a task that runs every minute:
    >>> @game.tasks.register(minutes=1, loop=True)
    >>> def example_task():
    >>>     ...
    """

    def __init__(self, game) -> None:
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.logger = logging.getLogger("tasks")
        self.tasks: List[Task] = []
        self.game = game

    def register(self, *, seconds=0, minutes=0, hours=0, loop=False, threaded=False):
        """Register a task

        `seconds`, `minutes`, `hours`: Specify when this task should run

        `loop`: If set to `True`, the task will loop itself in the specified time interval

        `threaded`: This will run the task inside a thread.
            Note that this is **not recommended**, since interactions with the game client were not designed to be done asyncronous.
            I will try to fix this issue in the future.
        """

        def wrapper(f: Callable):
            total = (seconds) + (minutes * 60) + (hours * 60 * 60)
            self.tasks.append(
                Task(
                    function=f,
                    interval=total,
                    loop=loop,
                    last_call=datetime.now(),
                    threaded=threaded,
                )
            )
            return f

        return wrapper

    def execute(self) -> None:
        """Execute every currently registered task"""
        if not self.game.bancho.connected:
            return

        self.logger.debug("Running tasks...")

        for task in self.tasks:
            last_call = (datetime.now() - task.last_call).total_seconds()

            if last_call >= task.interval:
                task.last_call = datetime.now()

                if not task.loop:
                    self.tasks.remove(task)

                try:
                    self.logger.debug(f"Trying to run task: '{task.function.__name__}'")

                    if not task.threaded:
                        task.function()
                        self.logger.debug(
                            f"Task '{task.function.__name__}' was successfully executed."
                        )
                        continue

                    # if (
                    #     task.current_thread is not None
                    #     and not task.current_thread.done()
                    # ):
                    # Thread is still running
                    # TODO

                    future = self.executor.submit(task.function)
                    future.add_done_callback(self._thread_callback)

                    self.logger.debug(
                        f"Task '{task.function.__name__}' was submitted to executor."
                    )

                except Exception as exc:
                    self.logger.error(
                        f"Failed to run '{task.function.__name__}' task: {exc}",
                        exc_info=exc,
                    )

        self.logger.debug("Done.")

    def _thread_callback(self, future: Future) -> None:
        ...  # TODO



from .client import BanchoClient
from .packets import Packets
from .constants import (
    PresenceFilter,
    ClientPackets,
    ServerPackets,
    ReplayAction,
    StatusAction,
    ButtonState,
    Privileges,
    LoginError,
    Grade,
    Mods,
    Mode,
)

from . import constants
from . import packets
from . import streams
from . import client



from typing import Callable, Optional, List

from .constants import Mode, Mods, RankingType, CommentTarget
from ..objects.score import ScoreResponse
from ..objects.comment import Comment
from ..game import Game

import requests
import logging


class WebAPI:

    """WebAPI
    -----------

    Functions:
        - `check_updates`
        - `connect`
        - `verify`
        - `get_session`
        - `get_backgrounds`
        - `get_friends`
        - `get_scores`
        - `get_star_rating`
        - `get_favourites`
        - `get_comments`
        - `post_comment`
        - `get_replay`
        - `get_avatar`
    """

    def __init__(self, game: Game) -> None:
        self.game = game

        self.session = requests.Session()
        self.session.headers = {"User-Agent": "osu!", "osu-version": self.game.version}

        self.logger = logging.getLogger(f"osu!api-{game.version}")

        self.url = f"https://osu.{self.game.server}"

    def connected_to_bancho(self, *args, **kwargs) -> Callable:
        def wrapper(f: Callable):
            if not self.game.bancho.connected:
                return

            return f(*args, **kwargs)

        return wrapper

    def check_updates(self) -> Optional[dict]:
        self.logger.info("Checking for updates...")

        response = self.session.get(
            "https://osu.ppy.sh/web/check-updates.php",
            params={
                "action": "check",
                "stream": self.game.stream.lower(),
                "time": self.game.time,
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to get updates ({response.status_code})")
            return None

        if "fallback" in response.text:
            self.logger.error(f'Failed to get updates: "{response.text}"')
            return None

        return response.json()

    def connect(self, retry=False) -> bool:
        """This will perform a request on `/web/bancho_connect.php`."""

        self.logger.info("Connecting to bancho...")

        response = self.session.get(
            f"{self.url}/web/bancho_connect.php",
            params={
                "v": self.game.version,
                "u": self.game.username,
                "h": self.game.password_hash,
                "fx": "fail",  # dotnet version
                "ch": str(self.game.client.hash),
                "retry": int(retry),
            },
        )

        if not response.ok:
            if self.game.server == "ppy.sh":
                self.logger.error(
                    f"Error on login: {requests.status_codes._codes[response.status_code][0].upper()}"
                )
                self.logger.warning("Ignoring...")
            return True

        if "error" in response.text:
            self.logger.error(
                f'Error on login: {response.text.removeprefix("error: ")}'
            )

            if "verify" in response.text:
                self.verify(str(self.game.client.hash))
                return False

        return True

    def verify(self, hash: str, exit_after: bool = True) -> None:
        """This will print out a url, where the user can verify this client."""
        self.logger.info("Verification required.")
        self.logger.info(
            f'{self.url}/p/verify?u={self.game.username.replace(" ", "%20")}&reason=bancho&ch={hash}'
        )
        self.logger.info("You only need to do this once.")

        if exit_after:
            exit(0)

    def get_session(self) -> requests.Response:
        """Perform a request on `/web/osu-session.php`.\n
        I don't know what this actually does.\n
        My guess is that it checks, if somebody is already online with this account.
        """  # TODO

        response = self.session.post(
            f"{self.url}/web/osu-session.php",
            files={
                "u": self.game.username,
                "h": self.game.password_hash,
                "action": "check",
            },
        )

        return response

    def get_backgrounds(self) -> Optional[dict]:
        """This will perform a request on `/web/osu-getseasonal.php`."""

        response = self.session.get(f"{self.url}/web/osu-getseasonal.php")

        if response.ok:
            return response.json()

    def get_friends(self) -> List[int]:
        """This will perform a request on `/web/osu-getfriends.php`."""

        response = self.session.get(
            f"{self.url}/web/osu-getfriends.php",
            params={"u": self.game.username, "h": self.game.password_hash},
        )

        if response.ok:
            return [int(id) for id in response.text.split("\n") if id.isdigit()]

        return []

    def get_scores(
        self,
        beatmap_checksum: str,
        beatmap_file: str,
        set_id: int,
        mode: Mode = Mode.Osu,
        mods: Optional[Mods] = Mods.NoMod,
        rank_type=RankingType.Top,
        skip_scores: bool = False,
    ) -> Optional[ScoreResponse]:
        """Get top scores for a beatmap

        - `beatmap_checksum`: MD5 hash of the beatmap file
        - `beatmap_file`: Filename of the beatmap
        - `set_id`: BeatmapsetId for the beatmap
        - `mode`: Specify a mode
        - `mods`: Filter by mods (`rank_type` must be set to `SelectedMod`)
        - `rank_type`: osu.api.constants.RankingTy
            - `Top`: Global Ranking
            - `SelectedMod`: Global Ranking (Selected Mods) (Supporter)
            - `Friends`: Friend Ranking (Supporter)
            - `Country`: Country Ranking (Supporter)
        """

        response = self.session.get(
            f"{self.url}/web/osu-osz2-getscores.php",
            params={
                "s": int(skip_scores),
                "vv": 4,  # ?
                "v": rank_type.value,
                "c": beatmap_checksum,
                "f": beatmap_file,
                "m": mode.value,
                "i": set_id,
                "mods": mods.value,
                "a": 0,  # ?
                "us": self.game.username,
                "ha": self.game.password_hash,
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to fetch scores ({response.status_code})")
            return

        return ScoreResponse.from_string(response.text, mode)

    def get_star_rating(
        self, beatmap_id: int, mode: Mode = Mode.Osu, mods: Mods = Mods.NoMod
    ) -> float:
        """Get star rating of a beatmap"""

        response = self.session.post(
            f"{self.url}/difficulty-rating",
            json={
                "beatmap_id": beatmap_id,
                "ruleset_id": mode.value,
                "mods": [{"acronym": acronym} for acronym in mods.acronyms],
            },
        )

        if not response.ok:
            return 0.0

        return float(response.text)

    def get_favourites(self) -> List[int]:
        """Get your beatmap favourites"""

        response = self.session.get(
            f"{self.url}/web/osu-getfavourites.php",
            params={"u": self.game.username, "h": self.game.password_hash},
        )

        if not response.ok:
            return []

        return [
            int(beatmap_id) for beatmap_id in response.text.split("\n") if beatmap_id
        ]

    def get_comments(
        self,
        beatmap_id: Optional[int] = None,
        set_id: Optional[int] = None,
        replay_id: Optional[int] = None,
        mode: Mode = Mode.Osu,
    ) -> List:
        """Get comments for a beatmap, set or replay"""

        if all([beatmap_id is None, set_id is None, replay_id is None]):
            return []

        response = self.session.post(
            f"{self.url}/web/osu-comment.php",
            files={
                "u": (None, self.game.username),
                "p": (None, self.game.password_hash),
                "b": (None, beatmap_id),
                "s": (None, set_id),
                "r": (None, replay_id),
                "m": (None, mode.value),
                "a": (None, "get"),
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to fetch comments ({response.status_code})")
            return []

        if not response.text:
            return []

        return [Comment.from_string(line) for line in response.text.split("\n") if line]

    def post_comment(
        self,
        text: str,
        time: int,
        target: CommentTarget = CommentTarget.Map,
        beatmap_id: Optional[int] = None,
        replay_id: Optional[int] = None,
        set_id: Optional[int] = None,
        mode: Mode = Mode.Osu,
    ) -> None:
        """Post a comment to a map, song or replay"""

        if all([beatmap_id is None, set_id is None, replay_id is None]):
            return

        target_id = {
            CommentTarget.Map: beatmap_id,
            CommentTarget.Song: set_id,
            CommentTarget.Replay: replay_id,
        }[target]

        if target_id is None:
            return

        self.session.post(
            f"{self.url}/web/osu-comment.php",
            files={
                "u": (None, self.game.username),
                "p": (None, self.game.password_hash),
                "b": (None, beatmap_id),
                "s": (None, set_id),
                "r": (None, replay_id),
                "m": (None, mode.value),
                "a": (None, "post"),
                "starttime": (None, time),
                "comment": (None, text),
                "target": (None, target.value),
            },
        )

    def get_replay(self, replay_id: int, mode: Mode = Mode.Osu) -> Optional[bytes]:
        """Get raw replay data by id (not osr!)"""

        response = self.session.get(
            f"{self.url}/web/osu-getreplay.php",
            params={
                "u": self.game.username,
                "h": self.game.password_hash,
                "m": mode.value,
                "c": replay_id,
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to fetch replay ({response.status_code})")
            return

        return response.content

    def get_avatar(self, user_id: int) -> Optional[bytes]:
        """Get avatar by user id"""

        return self.session.get(f"https://a.{self.game.server}/{user_id}").content

    # TODO: Add favourite
    # TODO: osu! direct
    # TODO: osz download



from ..bancho.streams import StreamIn
from ..bancho.constants import Grade

from dataclasses import dataclass


@dataclass
class BeatmapInfo:
    id: int
    beatmap_id: int
    beatmapset_id: int
    thread_id: int
    ranked: int
    osu_rank: Grade
    fruits_rank: Grade
    taiko_rank: Grade
    mania_rank: Grade
    checksum: str

    @classmethod
    def decode(cls, stream: StreamIn):
        return BeatmapInfo(
            stream.s16(),
            stream.s32(),
            stream.s32(),
            stream.s32(),
            stream.u8(),
            Grade(stream.u8()),
            Grade(stream.u8()),
            Grade(stream.u8()),
            Grade(stream.u8()),
            stream.string(),
        )



import setuptools
import os

current_directory = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(current_directory, "README.md"), "r") as f:
    long_description = f.read()

with open(os.path.join(current_directory, "requirements.txt"), "r") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="osu",
    version="1.2.3",
    author="Lekuru",
    author_email="contact@lekuru.xyz",
    description="A python library that emulates the osu! stable client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    keywords=["osu", "osugame", "python", "bancho"],
)



from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Callable, Dict, List
from datetime import datetime
from copy import copy

from .bancho.constants import ServerPackets
from .objects.client import ClientInfo
from .tasks import Task

import requests
import hashlib
import logging


class Game:

    """osu! game client
    --------------------
    This class will try to emulate the online functionality of the osu! stable client.
    """

    def __init__(
        self,
        username: str,
        password: str,
        server="ppy.sh",
        stream="stable40",
        version: Optional[int] = None,
        tournament=False,
        events: Optional[Dict[ServerPackets, List[Callable]]] = {},
        tasks: Optional[List[Task]] = [],
    ) -> None:
        """Parameters
        -------------

        `username`: str
            Your bancho username

        `password`: str
            Your bancho password

        `server`: str
            Server endpoint
            (`ppy.sh`)

        `stream`: str
            Release stream
            (`stable40`)

        `version`: int, optional
            Set a custom version
            (Will fetch the latest version automatically)

        `tournament`: bool
            This allows for multiple clients at the same time (supporter only)

        `events`: dict, optional
            Allows to pass in events, instead of registering them normally

        `tasks`: list, optional
            Allows to pass in pre-defined tasks, just like the `events` parameter
        """

        self.username = username
        self.password = password
        self.server = server
        self.stream = stream
        self.version = version
        self.tourney = tournament

        self.version_number = version

        self.logger = logging.getLogger("osu!")

        if not version:
            # Fetch latest client version
            self.version = self.fetch_version(stream)

            if not self.version:
                # Failed to get version
                exit(1)
        else:
            # Custom client version was set
            self.version = f"b{self.version}"

            if self.tourney:
                self.version = f"{self.version}tourney"

        self.session = requests.Session()
        self.session.headers = {"User-Agent": "osu!", "osu-version": self.version}

        self.logger.name = f"osu!-{self.version}"

        from .bancho import BanchoClient, Packets
        from .events import EventHandler
        from .tasks import TaskManager
        from .api import WebAPI

        self.packets = copy(Packets)
        self.events = EventHandler()
        self.bancho = BanchoClient(self)
        self.tasks = TaskManager(self)
        self.api = WebAPI(self)

        if events:
            self.events.handlers = events

        if tasks:
            self.tasks.tasks = tasks

        if not (updates := self.api.check_updates()):
            # Updates are required because of the executable hash
            # TODO: Custom executable hash?
            exit(1)

        self.client = ClientInfo(self.version, updates)

    def __repr__(self) -> str:
        return f"<osu! {self.version}>"

    @property
    def password_hash(self) -> str:
        """Return md5 hashed password"""
        return hashlib.md5(self.password.encode()).hexdigest()

    @property
    def time(self) -> int:
        """Return current time as ticks"""
        return int((datetime.now() - datetime(1, 1, 1)).total_seconds() * 10000000)

    def run(self, retry=False, exit_on_interrupt=False) -> None:
        if retry:
            # Reinitialize game
            self.__init__(
                self.username,
                self.password,
                self.server,
                self.stream,
                self.version_number,
                self.tourney,
                self.events.handlers,
                self.tasks.tasks,
            )

        try:
            if not retry:
                self.api.get_session()
                self.api.get_backgrounds()

            if not self.api.connect(retry):
                exit(1)

            self.bancho.run()

        except KeyboardInterrupt:
            pass

        finally:
            self.logger.warning("Exiting...")
            self.bancho.exit()
            self.logger.warning("Stopping tasks...")
            self.events.executor.shutdown()
            self.tasks.executor.shutdown()

        if exit_on_interrupt:
            exit(0)

    def fetch_version(self, stream: str = "stable40") -> Optional[str]:
        """
        Fetch the latest version of the client from:
        <https://osu.ppy.sh/home/changelog>
        """

        self.logger.info("Fetching client version...")

        response = requests.get(
            f"https://osu.ppy.sh/home/changelog/{stream}",
            allow_redirects=True,
            headers={"User-Agent": "osu!"},
        )

        if not response.ok:
            self.logger.error(f"Failed to get client version ({response.status_code}).")
            self.logger.error(
                "Please check your release stream or provide a custom version!"
            )
            return None

        # Parse url (e.g.: https://osu.ppy.sh/home/changelog/stable40/20230326)
        version = response.url.removesuffix("/").split("/")[-1]

        if not version.replace(".", "").isdigit():
            self.logger.error(
                f'Failed to get client version ("{version}" is not an integer).'
            )
            self.logger.error("Please provide a custom version!")
            return None

        self.logger.debug(f"Version: b{version}{'tourney' if self.tourney else ''}")

        try:
            self.version_number = int(version)
        except ValueError:
            self.version_number = float(version)

        return f"b{version}{'tourney' if self.tourney else ''}"



from enum import IntEnum, Enum, IntFlag
from typing import List


class ClientPackets(IntEnum):
    CHANGE_ACTION = 0
    SEND_PUBLIC_MESSAGE = 1
    LOGOUT = 2
    REQUEST_STATUS_UPDATE = 3
    PING = 4
    START_SPECTATING = 16
    STOP_SPECTATING = 17
    SPECTATE_FRAMES = 18
    ERROR_REPORT = 20
    CANT_SPECTATE = 21
    SEND_PRIVATE_MESSAGE = 25
    PART_LOBBY = 29
    JOIN_LOBBY = 30
    CREATE_MATCH = 31
    JOIN_MATCH = 32
    PART_MATCH = 33
    MATCH_CHANGE_SLOT = 38
    MATCH_READY = 39
    MATCH_LOCK = 40
    MATCH_CHANGE_SETTINGS = 41
    MATCH_START = 44
    MATCH_SCORE_UPDATE = 47
    MATCH_COMPLETE = 49
    MATCH_CHANGE_MODS = 51
    MATCH_LOAD_COMPLETE = 52
    MATCH_NO_BEATMAP = 54
    MATCH_NOT_READY = 55
    MATCH_FAILED = 56
    MATCH_HAS_BEATMAP = 59
    MATCH_SKIP_REQUEST = 60
    CHANNEL_JOIN = 63
    BEATMAP_INFO_REQUEST = 68
    MATCH_TRANSFER_HOST = 70
    FRIEND_ADD = 73
    FRIEND_REMOVE = 74
    MATCH_CHANGE_TEAM = 77
    CHANNEL_PART = 78
    RECEIVE_UPDATES = 79
    SET_AWAY_MESSAGE = 82
    IRC_ONLY = 84
    USER_STATS_REQUEST = 85
    MATCH_INVITE = 87
    MATCH_CHANGE_PASSWORD = 90
    TOURNAMENT_MATCH_INFO_REQUEST = 93
    USER_PRESENCE_REQUEST = 97
    USER_PRESENCE_REQUEST_ALL = 98
    TOGGLE_BLOCK_NON_FRIEND_DMS = 99
    TOURNAMENT_JOIN_MATCH_CHANNEL = 108
    TOURNAMENT_LEAVE_MATCH_CHANNEL = 109

    def __repr__(self) -> str:
        return f"<{self.name} ({self.value})>"


class ServerPackets(IntEnum):
    USER_ID = 5
    SEND_MESSAGE = 7
    PONG = 8
    HANDLE_IRC_CHANGE_USERNAME = 9  # unused
    HANDLE_IRC_QUIT = 10
    USER_STATS = 11
    USER_LOGOUT = 12
    SPECTATOR_JOINED = 13
    SPECTATOR_LEFT = 14
    SPECTATE_FRAMES = 15
    VERSION_UPDATE = 19
    SPECTATOR_CANT_SPECTATE = 22
    GET_ATTENTION = 23
    NOTIFICATION = 24
    UPDATE_MATCH = 26
    NEW_MATCH = 27
    DISPOSE_MATCH = 28
    TOGGLE_BLOCK_NON_FRIEND_DMS = 34
    MATCH_JOIN_SUCCESS = 36
    MATCH_JOIN_FAIL = 37
    FELLOW_SPECTATOR_JOINED = 42
    FELLOW_SPECTATOR_LEFT = 43
    ALL_PLAYERS_LOADED = 45
    MATCH_START = 46
    MATCH_SCORE_UPDATE = 48
    MATCH_TRANSFER_HOST = 50
    MATCH_ALL_PLAYERS_LOADED = 53
    MATCH_PLAYER_FAILED = 57
    MATCH_COMPLETE = 58
    MATCH_SKIP = 61
    UNAUTHORIZED = 62  # unused
    CHANNEL_JOIN_SUCCESS = 64
    CHANNEL_INFO = 65
    CHANNEL_KICK = 66
    CHANNEL_AUTO_JOIN = 67
    BEATMAP_INFO_REPLY = 69
    PRIVILEGES = 71
    FRIENDS_LIST = 72
    PROTOCOL_VERSION = 75
    MAIN_MENU_ICON = 76
    MONITOR = 80  # unused
    MATCH_PLAYER_SKIPPED = 81
    USER_PRESENCE = 83
    RESTART = 86
    MATCH_INVITE = 88
    CHANNEL_INFO_END = 89
    MATCH_CHANGE_PASSWORD = 91
    SILENCE_END = 92
    USER_SILENCED = 94
    USER_PRESENCE_SINGLE = 95
    USER_PRESENCE_BUNDLE = 96
    USER_DM_BLOCKED = 100
    TARGET_IS_SILENCED = 101
    VERSION_UPDATE_FORCED = 102
    SWITCH_SERVER = 103
    ACCOUNT_RESTRICTED = 104
    RTX = 105  # unused
    MATCH_ABORT = 106
    SWITCH_TOURNAMENT_SERVER = 107

    def __repr__(self) -> str:
        return f"<{self.name} ({self.value})>"


class LoginError(Enum):
    AUTHENTICATION_ERROR = -1
    UPDATE_NEEDED = -2
    RESTRICTED = -3
    NOT_ACTIVATED = -4
    SERVER_ERROR = -5
    NEED_SUPPORTER = -6
    PASSWORD_RESET = -7
    VERIFICATION_NEEDED = -8

    @property
    def description(self) -> str:
        return {
            LoginError.AUTHENTICATION_ERROR: "Authentication failed. Please check your username/password!",
            LoginError.UPDATE_NEEDED: "It seems like this version of osu! is too old. Please check for any updates!",
            LoginError.RESTRICTED: "You are banned.",
            LoginError.NOT_ACTIVATED: "Your account was either restricted or is not activated.",
            LoginError.SERVER_ERROR: "A server error occured.",
            LoginError.NEED_SUPPORTER: "You need to be a supporter to use tourney clients.",
            LoginError.PASSWORD_RESET: "Your account password has been reset.",
            LoginError.VERIFICATION_NEEDED: "",
        }[self]


class StatusAction(Enum):
    Idle = 0
    Afk = 1
    Playing = 2
    Editing = 3
    Modding = 4
    Multiplayer = 5
    Watching = 6
    Unknown = 7
    Testing = 8
    Submitting = 9
    Paused = 10
    Lobby = 11
    Multiplaying = 12
    OsuDirect = 13


class Mods(IntFlag):
    NoMod = 0
    NoFail = 1 << 0
    Easy = 1 << 1
    NoVideo = 1 << 2
    Hidden = 1 << 3
    HardRock = 1 << 4
    SuddenDeath = 1 << 5
    DoubleTime = 1 << 6
    Relax = 1 << 7
    HalfTime = 1 << 8
    Nightcore = 1 << 9
    Flashlight = 1 << 10
    Autoplay = 1 << 11
    SpunOut = 1 << 12
    Autopilot = 1 << 13
    Perfect = 1 << 14
    Key4 = 1 << 15
    Key5 = 1 << 16
    Key6 = 1 << 17
    Key7 = 1 << 18
    Key8 = 1 << 19
    FadeIn = 1 << 20
    Random = 1 << 21
    Cinema = 1 << 22
    Target = 1 << 23
    Key9 = 1 << 24
    KeyCoop = 1 << 25
    Key1 = 1 << 26
    Key3 = 1 << 27
    Key2 = 1 << 28
    ScoreV2 = 1 << 29
    LastMod = 1 << 30
    ScoreIncreaseMods = Hidden | HardRock | DoubleTime | Flashlight | FadeIn
    KeyMod = Key1 | Key2 | Key3 | Key4 | Key5 | Key6 | Key7 | Key8 | Key9 | KeyCoop
    FreeModAllowed = (  # black coding style be like
        NoFail
        | Easy
        | Hidden
        | HardRock
        | SuddenDeath
        | Flashlight
        | FadeIn
        | Relax
        | Autopilot
        | SpunOut
        | KeyMod
    )

    @property
    def members(self) -> list:
        return [flag for flag in Mods if self & flag]

    @property
    def acronyms(self) -> List[str]:
        return [
            {
                Mods.NoFail: "NF",
                Mods.Easy: "EZ",
                Mods.Hidden: "HD",
                Mods.HardRock: "HR",
                Mods.SuddenDeath: "SD",
                Mods.DoubleTime: "DT",
                Mods.Relax: "Relax",
                Mods.HalfTime: "HT",
                Mods.Nightcore: "NC",
                Mods.Flashlight: "FL",
                Mods.SpunOut: "SO",
                Mods.Autopilot: "AP",
                Mods.Perfect: "PF",
                Mods.Key1: "K1",
                Mods.Key2: "K2",
                Mods.Key3: "K3",
                Mods.Key4: "K4",
                Mods.Key5: "K5",
                Mods.Key6: "K6",
                Mods.Key7: "K7",
                Mods.Key8: "K8",
                Mods.KeyCoop: "2P",
                Mods.FadeIn: "FI",
                Mods.Random: "RD",
                Mods.ScoreV2: "ScoreV2",
                Mods.Cinema: "Cinema",
                Mods.Autoplay: "Auto",
                Mods.Target: "TP",
            }[mod]
            for mod in self.members
            if mod
            not in [
                Mods.ScoreIncreaseMods,
                Mods.FreeModAllowed,
                Mods.LastMod,
                Mods.KeyMod,
                Mods.NoMod,
            ]
        ]


class Privileges(IntFlag):
    Restriced = 0
    Normal = 1
    BAT = 2
    Supporter = 4
    Peppy = 8
    Admin = 16
    Tournament = 32


class ButtonState(IntFlag):
    NoButtons = 0
    Left1 = 1
    Right1 = 2
    Left2 = 4
    Right2 = 8
    Smoke = 16


class Mode(Enum):
    Osu = 0
    Taiko = 1
    CatchTheBeat = 2
    OsuMania = 3


class ReplayAction(Enum):
    Standard = 0
    NewSong = 1
    Skip = 2
    Completion = 3
    Fail = 4
    Pause = 5
    Unpause = 6
    SongSelect = 7
    WatchingOther = 8


class PresenceFilter(Enum):
    NoPlayers = 0
    All = 1
    Friends = 2


class Grade(Enum):
    XH = 0
    SH = 1
    X = 2
    S = 3
    A = 4
    B = 5
    C = 6
    D = 7
    F = 8
    N = 9



from dataclasses import dataclass, field
from typing import Any, List

from ..bancho.constants import StatusAction, Mode, Mods


@dataclass
class Status:
    action: StatusAction = StatusAction.Idle
    text: str = ""
    checksum: str = ""
    mods: Mods = Mods.NoMod
    mode: Mode = Mode.Osu
    beatmap_id: int = 0

    def __repr__(self) -> str:
        return f'<{self.action.name}{f" - {self.text}" if self.text else ""}>'

    def reset(self) -> None:
        self.action = StatusAction.Idle
        self.text = ""
        self.checksum = ""
        self.mods = []
        self.mode = Mode.Osu
        self.beatmap_id = 0



from ..bancho.constants import Mods, Mode

from enum import IntEnum, Enum


class RankingType(IntEnum):
    Top = 1
    SelectedMod = 2
    Friends = 3
    Country = 4


class SubmissionStatus(IntEnum):
    Unknown = 0
    NotSubmitted = 1
    Pending = 2
    EditableCutoff = 3
    Ranked = 4
    Approved = 5
    Qualified = 6


class CommentTarget(str, Enum):
    Map = "map"
    Song = "song"
    Replay = "replay"



from typing import List, Optional
from dateutil.tz import tzlocal
from datetime import datetime

import platform
import hashlib
import psutil


class ClientHash:

    """The client hash is used to identify a device on the server side."""

    def __init__(self, executable_hash: str) -> None:
        """
        Args:
            `executable_hash`: str
                The hash of your `osu!.exe`.
                Can also be retrieved from `/web/check-updates.php`

                Warning: Bancho will refuse the connection, if your version is too old!
        """

        self.executable_hash = executable_hash

    def __repr__(self) -> str:
        return f"{self.executable_hash}:{self.adapter_string}:{self.adapter_hash}:{self.uninstall_id}:{self.disk_signature}:"

    @property
    def adapters(self) -> List[str]:
        return [
            psutil.net_if_addrs()[adapter][0].address
            for adapter in psutil.net_if_addrs()
        ]

    @property
    def adapter_string(self) -> str:
        adapters = [
            adapter.replace("-", "")
            for adapter in self.adapters
            if adapter.count("-") == 5
        ]
        adapters.insert(3, "")
        return ".".join(adapters)

    @property
    def adapter_hash(self) -> str:
        if platform.system() != "Windows":
            return "runningunderwine"

        return hashlib.md5(self.adapter_string.encode()).hexdigest()

    @property
    def uninstall_id(self) -> str:
        if platform.system() != "Windows":
            return hashlib.md5(b"unknown").hexdigest()

        import winreg

        try:
            # Try to read the UninstallID from Registry
            with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
                key = winreg.OpenKey(reg, "Software\\osu!")

                for i in range(1024):
                    name, value, type = winreg.EnumValue(key, i)

                    if name != "UninstallID":
                        continue

                    return hashlib.md5(value.encode()).hexdigest()
        except (FileNotFoundError, WindowsError, EnvironmentError):
            # Key was not found
            pass

        return hashlib.md5(b"unknown").hexdigest()

    @property
    def disk_signature(self) -> str:
        if platform.system() != "Windows":
            return hashlib.md5(b"unknown").hexdigest()

        from wmi import WMI

        # Get serial number of first item
        for item in WMI().query("SELECT * FROM Win32_DiskDrive"):
            return hashlib.md5(item.SerialNumber.encode()).hexdigest()

        # Fallback
        return hashlib.md5(b"unknown").hexdigest()


class ClientInfo:

    """Sent when logging in to bancho.

    Parameters:
        `version`: str
            Client version (e.g. b20220829)

        `utc_offset`: int
            UTC offset for this device

        `display_city`: bool
            Probably has something to do with the world map inside osu!

        `hash`: osu.objects.client.ClientHash
            See above

        `friendonly_dms`: bool
            aka. "Block non-friend dms"
    """

    def __init__(self, version: str, updates: dict) -> None:
        """
        Args:
            `version`: str
                Client version (e.g. b20220829)

            `updates`: dict
                Update response from `/web/check-updates.php`
        """

        self.hash = ClientHash(self.get_file_hash(updates))
        self.version = version

        self.friendonly_dms = False
        self.display_city = False

        self.utc_offset = round(
            datetime.now(tzlocal()).utcoffset().total_seconds() / 3600
        )

    def __repr__(self) -> str:
        return f"{self.version}|{self.utc_offset}|{int(self.display_city)}|{self.hash}|{int(self.friendonly_dms)}"

    @classmethod
    def get_file_hash(cls, updates: dict) -> Optional[str]:
        for file in updates:
            if file["filename"] == "osu!.exe":
                return file["file_hash"]
        return None



from typing import Optional, List
from datetime import datetime
from queue import Queue

from .constants import ClientPackets, ReplayAction, StatusAction, Privileges
from .streams import StreamOut

from ..objects.replays import ReplayFrame, ScoreFrame
from ..objects.collections import Players, Channels
from ..objects.player import Player
from ..objects.status import Status

from ..game import Game

import requests
import logging
import time


class BanchoClient:

    """BanchoClient
    ---------------

    Parameters:
        `player`: osu.objects.Player

        `spectating`: osu.objects.Player

        `players`: osu.objects.Players

        `channels`: osu.objects.Channels

        `privileges`: osu.bancho.constants.Privileges

        `friends`: list[int]

        `game`: osu.game.Game

        `connected`: bool

        `retry`: bool

    Functions:
        `enqueue`: Send a bancho packet to the server

        `ping`: Send a ping packet

        `request_presence`: Request a presence update for a list of players

        `request_stats`: Request a stats update for a list of players

        `request_status`: Request a status update for your account

        `update_status`: Update `player.status`

        `start_spectating`: Start spectating a player

        `stop_spectating`: Stop spectating

        `cant_spectate`: Is sent when the player does not have the beatmap for spectating

        `send_frames`: Send replay frames to your spectators

        `join_lobby`: Join the lobby

        `leave_lobby`: Leave the lobby
    """

    def __init__(self, game: Game) -> None:
        self.game = game

        self.logger = logging.getLogger(f"bancho-{game.version}")

        self.domain = f"c.{game.server}"
        self.url = f"https://{self.domain}"

        self.session = requests.Session()
        self.session.headers = {
            "osu-version": self.game.version,
            "Accept-Encoding": "gzip, deflate",
            "User-Agent": "osu!",
            "Host": self.domain,
        }

        self.user_id = -1
        self.connected = False
        self.retry = True
        self.token = ""

        self.spectating: Optional[Player] = None
        self.player: Optional[Player] = None

        self.channels = Channels()
        self.players = Players(game)
        self.queue = Queue()

        self.ping_count = 0
        self.protocol = 0

        self.privileges: Privileges = Privileges.Normal
        self.friends: List[int] = []

        self.last_action = datetime.now().timestamp()
        self.fast_read = False

        self.silenced = False
        self.in_lobby = False

        self.min_idletime = 1
        self.max_idletime = 4

    @property
    def status(self) -> Status:
        if not self.player:
            return Status()
        return self.player.status

    @property
    def idle_time(self) -> float:
        """Time between now and the last request"""
        return datetime.now().timestamp() - self.last_action

    @property
    def request_interval(self) -> int:
        """Time between requests"""

        if self.fast_read:
            return 0

        interval = 1

        if self.game.tourney:
            return interval

        if not self.spectating:
            interval *= 1 + self.idle_time / 10
            interval *= 1 + self.ping_count

        interval = min(self.max_idletime, max(self.min_idletime, interval))

        return interval

    def run(self):
        self.connect()

        while self.connected:
            try:
                self.dequeue()
                self.game.tasks.execute()
                time.sleep(self.request_interval)
            except KeyboardInterrupt:
                raise
            except Exception as exc:
                self.logger.fatal(f"Unhandled Exception: {exc}", exc_info=exc)

        if self.retry:
            self.logger.error("Retrying in 15 seconds...")
            time.sleep(15)
            self.game.run(retry=True)

    def connect(self):
        data = f"{self.game.username}\n{self.game.password_hash}\n{self.game.client}\n"

        response = self.session.post(self.url, data=data)

        if not response.ok:
            self.connected = False
            self.retry = True
            self.logger.error(
                f"[{response.url}]: Connection was refused ({response.status_code})"
            )
            return

        if not (token := response.headers.get("cho-token")):
            # Failed to get token
            self.connected = False
            self.retry = False
            self.game.packets.data_received(response.content, self.game)
            return

        self.connected = True
        self.token = token

        self.session.headers["osu-token"] = self.token

        self.game.packets.data_received(response.content, self.game)

    def dequeue(self) -> Optional[requests.Response]:
        """Send a request to bancho, empty the queue and handle incoming packets"""

        if not self.connected:
            return

        if self.queue.empty():
            # Queue is empty, sending ping
            self.ping_count += 1
            return self.ping()
        else:
            self.ping_count = 0

        data = b""

        while not self.queue.empty():
            data += self.queue.get()

        response = self.session.post(self.url, data=data)

        if not response.ok:
            self.connected = False
            self.retry = True
            self.logger.error(
                f"[{response.request.url}]: Connection was refused ({response.status_code})"
            )
            return

        self.fast_read = False
        self.game.packets.data_received(response.content, self.game)

        self.last_action = datetime.now().timestamp()

        return response

    def exit(self):
        """Send logout packet to bancho, and disconnect."""

        self.enqueue(ClientPackets.LOGOUT, int(0).to_bytes(4, "little"))
        self.connected = False
        self.retry = False

    def enqueue(self, packet: ClientPackets, data: bytes = b"", dequeue=True) -> bytes:
        """Send a packet to the queue and dequeue"""

        stream = StreamOut()

        # Construct header
        stream.u16(packet.value)
        stream.bool(False)
        stream.u32(len(data))
        stream.write(data)

        self.logger.debug(f'Sending {packet.name} -> "{data}"')

        # Append to queue
        self.queue.put(stream.get())

        if dequeue:
            self.dequeue()

        return stream.get()

    def unsilence(self):
        if not self.silenced:
            return

        self.player.silenced = False
        self.silenced = False

        self.logger.info("You can now talk again.")

    def ping(self):
        self.enqueue(ClientPackets.PING)

    def request_presence(self, ids: List[int]):
        stream = StreamOut()
        stream.intlist(ids)

        self.enqueue(ClientPackets.USER_PRESENCE_REQUEST, stream.get())

    def request_stats(self, ids: List[int]):
        stream = StreamOut()
        stream.intlist(ids)

        self.enqueue(ClientPackets.USER_STATS_REQUEST, stream.get())

    def request_status(self):
        self.enqueue(ClientPackets.REQUEST_STATUS_UPDATE)

    def update_status(self):
        """Send current player status to the server

        You can change your status by updating the `game.bancho.player.status` attributes.\n
        For example:
            >>> status = game.bancho.player.status
            >>>
            >>> status.action = StatusAction.Playing
            >>> status.text = "UNDEAD CORPORATION - Everything will freeze [Time Freeze]"
            >>> status.checksum = "a84050da9b68ca1bd8e2d1700b9c6ca5" # .osu file checksum
            >>> status.mods = Mods.Hidden|Mods.HardRock
            >>> status.mode = GameMode.Osu
            >>> status.beatmap_id = 555797
            >>>
            >>> game.bancho.update_status()
        """

        if not self.player:
            return

        stream = StreamOut()
        stream.u8(self.player.status.action.value)
        stream.string(self.player.status.text)
        stream.string(self.player.status.checksum)
        stream.u32(self.player.status.mods.value)
        stream.u8(self.player.mode.value)
        stream.s32(self.player.status.beatmap_id)

        self.enqueue(ClientPackets.CHANGE_ACTION, stream.get())

    def start_spectating(self, target: Player):
        """Start spectating a player

        You will receive `ServerPackets.SPECTATE_FRAMES` packets that contain replay data.\n
        You can use osu-recorder (https://github.com/Lekuruu/osu-recorder) as a reference for that.
        """
        if self.spectating:
            self.stop_spectating()

        self.enqueue(
            ClientPackets.START_SPECTATING,
            int(target.id).to_bytes(4, "little"),
            dequeue=False,
        )

        self.spectating = target

        target.request_presence()
        target.request_stats()

        self.status.action = StatusAction.Watching
        self.status.text = target.status.text
        self.status.checksum = target.status.checksum
        self.status.mods = target.status.mods
        self.status.mode = target.status.mode
        self.status.beatmap_id = target.status.beatmap_id

        self.update_status()

    def stop_spectating(self):
        if not self.spectating:
            return

        self.logger.info(f"Stopped spectating {self.spectating.name}.")

        self.enqueue(ClientPackets.STOP_SPECTATING, dequeue=False)

        self.status.reset()
        self.update_status()

    def cant_spectate(self):
        self.enqueue(ClientPackets.CANT_SPECTATE)

    def send_frames(
        self,
        action: ReplayAction,
        frames: List[ReplayFrame],
        score_frame: Optional[ScoreFrame] = None,
        seed: int = 0,
    ):
        """
        This will send a `SPECTATE_FRAMES` packet to the server, which gets broadcasted to all of your spectators.
        Note that this will not work, if nobody is spectating you.
        """
        if not self.player.spectators:
            self.logger.warning(
                "Failed to send frames, because no spectators were found."
            )
            return

        if self.spectating:
            action = ReplayAction.WatchingOther
            extra = self.spectating.id
        else:
            extra = seed

        stream = StreamOut()
        stream.s32(extra)
        stream.u16(len(frames))
        [stream.write(frame.encode()) for frame in frames]
        stream.u8(action.value)

        if score_frame:
            stream.write(score_frame.encode())

        self.enqueue(ClientPackets.SPECTATE_FRAMES, stream.get())

    def join_lobby(self):
        if self.in_lobby:
            return

        self.enqueue(ClientPackets.JOIN_LOBBY)
        self.in_lobby = True

    def leave_lobby(self):
        if not self.in_lobby:
            return

        self.enqueue(ClientPackets.PART_LOBBY)
        self.in_lobby = True

    def set_away_message(self, message: str):
        raise NotImplementedError  # TODO



from typing import Iterable, Iterator, Set, Any
from threading import Lock


class LockedSet(set):
    """A set where add(), remove(), and 'in' operator are thread-safe"""

    def __init__(self, *args, **kwargs):
        self._lock = Lock()
        super(LockedSet, self).__init__(*args, **kwargs)

    def __len__(self) -> int:
        with self._lock:
            return super(LockedSet, self).__len__()

    def __iter__(self) -> Iterator:
        with self._lock:
            return super(LockedSet, self).__iter__()

    def __or__(self, __value: Set) -> Set[Any]:
        with self._lock:
            return super(LockedSet, self).__or__(__value)

    def __contains__(self, obj: object) -> bool:
        with self._lock:
            return super(LockedSet, self).__contains__(obj)

    def add(self, element: Any) -> None:
        with self._lock:
            return super(LockedSet, self).add(element)

    def remove(self, element: Any) -> None:
        with self._lock:
            super(LockedSet, self).remove(element)

    def update(self, *s: Iterable) -> None:
        with self._lock:
            super(LockedSet, self).update(*s)
